#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.
"""Application wide settings that are shared across multiple modules"""

import os
import sys
from distutils.util import strtobool

from dotenv import load_dotenv

from constants import ONE_MB, DEFAULT_GAIN, DEFAULT_IT_TIME

try:
    load_dotenv()

    # Logging
    log_file = os.getenv('log-file', '/home/pi/rcm-station-client.log')
    log_level = os.getenv('log-level', 'WARN')
    log_max_bytes = int(os.getenv('log-max-bytes', ONE_MB))
    log_backup_count = int(os.getenv('log-backup-count', '5'))
    # pylint: disable=invalid-name
    log_include_console_handler = bool(strtobool(os.getenv('log-include-console-handler', 'false')))

    # Requests
    timeout = int(os.getenv('requests-timeout', '5'))
    verify = False

    # RCM specific
    api_base_url = os.getenv('api-base-url', 'https://raspberrypi.local:8080')
    reports_endpoint = api_base_url + '/measurement-reports'
    stations_endpoint = api_base_url + '/stations'
    station_name = os.getenv('station-name')
    sensors = os.getenv('sensors')

    tsl2591_gain = os.getenv('tsl2591-gain', DEFAULT_GAIN)
    tsl2591_integration_time = os.getenv('tsl2591-integration-time', DEFAULT_IT_TIME)
    # pylint: disable=invalid-name
    sgp30_init_baseline = bool(strtobool(os.getenv('sgp30-init-baseline', 'false')))
    sgp30_baseline_file = os.getenv('sgp30-baseline-file', '/home/pi/rcm-sgp30-baseline.txt')

    # Security
    # pylint: disable=invalid-name
    auth_enabled = bool(strtobool(os.getenv('auth-enabled', 'true')))
    token_endpoint = os.getenv('auth-token-endpoint')
    client_id = os.getenv('auth-client-id', 't6N95EF5uZgXJYqYONKXmiwTMSwxpcvw')
    client_secret = os.getenv('auth-client-secret')
    audience = os.getenv('auth-audience', 'https://station-monitoring-service')
    token_file = os.getenv('token-file', 'token.txt')

    # Other
    # pylint: disable=invalid-name
    disable_I2C = bool(strtobool(os.getenv('disable-i2c', 'false')))

    # Validation
    if station_name is None:
        raise ValueError('Station name is missing. Please configure "station-name" in the .env file')
    if sensors is None:
        raise ValueError('No sensors configured. Please configure "sensors" in the .env file')
    if auth_enabled and (token_endpoint is None or client_id is None or client_secret is None or audience is None):
        raise ValueError('Auth is not configured properly, please double check settings')
except Exception as exception:
    print('Invalid configuration', exception)
    sys.exit(1)



#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rcm-station-client",
    version="0.3.1",
    author="Pim Hazebroek",
    author_email="rcm@pimhazebroek.nl",
    description="Collects measurement reports from various sensors and send them to the station-monitoring-service.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/residential-climate-monitoring/station-client",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'requests==2.25.0',
        'urllib3==1.26.2',
        'python-dotenv==0.15.0',
        'psutil==5.7.3',
        'RPI.GPIO==0.7.0',
        'adafruit-blinka==5.8.0',
        'adafruit-circuitpython-busdevice==5.0.1',
        'adafruit-circuitpython-sht31d==2.3.2',
        'adafruit-circuitpython-bme680==3.2.4',
        'adafruit-circuitpython-tsl2591==1.2.4',
        'adafruit-circuitpython-sgp30==2.3.1',
        # Remember to sync dependencies in requirements.txt
    ],
    extras_require={
        'build': [
            'setuptools==50.3.2',
            'pylint',
            'bump2version==1.0.0'
        ]
    },
    python_requires='>=3.7',
)



#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.

"""
The SHT31-D sensor - High precision Temperature & Humidity sensor
=================================================================

See https://www.adafruit.com/product/2857
"""
import logging
from distutils.util import strtobool

import adafruit_sht31d

from constants import SENSOR_SHT31_D
from sensors.sensor import Sensor

_PRIMARY_ADDRESS = 0x44
_SECONDARY_ADDRESS = 0x45


class SHT31D(Sensor):
    """Concrete implementation of SHT31-D sensor."""

    @property
    def name(self):
        return SENSOR_SHT31_D

    def read(self):
        """
        Reads the temperature and relative humidity.
        """
        logging.info('Reading SHT31-D sensor')
        address = self._get_address()
        sensor = adafruit_sht31d.SHT31D(self._i2c, address)
        measurements = {
            'temperature': sensor.temperature,
            'humidity': sensor.relative_humidity
        }
        logging.info('Measurement: %s', measurements)
        return measurements

    def _get_address(self):
        """Determine the I2C address to use"""
        if 'secondary_address' not in self._settings.keys():
            address = _PRIMARY_ADDRESS
        else:
            address = _SECONDARY_ADDRESS if strtobool(self._settings['secondary_address']) else _PRIMARY_ADDRESS

        return address



#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.

"""
The TSL2591 sensor - High Dynamic Range Digital Light Sensor
============================================================

See https://www.adafruit.com/product/439

Gain & Timing
You can adjust the gain settings and integration time of the sensor to make it more or less sensitive to light, depending on the
environment where the sensor is being used.

Gain:
GAIN_LOW: Sets the gain to 1x (bright light)
GAIN_MEDIUM or GAIN_MED: Sets the gain to 25x (general purpose)
GAIN_HIGH: Sets the gain to 428x (low light)
GAIN_MAX: Sets the gain to 9876x (extremely low light)

The integration time can be set between 100 and 600ms, and the longer the integration time the more light the sensor is able to integrate,
making it more sensitive in low light the longer the integration time. The following values can be used:
100MS
200MS
300MS
400MS
500MS
600MS
"""
import logging

import adafruit_tsl2591

from constants import GAIN_LOW, GAIN_MEDIUM, GAIN_HIGH, GAIN_MAX
from constants import IT_TIME_100MS, IT_TIME_200MS, IT_TIME_300MS, IT_TIME_400MS, IT_TIME_500MS, IT_TIME_600MS, SENSOR_TSL2591
from sensors.sensor import Sensor
from settings import tsl2591_gain, tsl2591_integration_time

# Maximum lux the sensor can measure
_MAX_LUX = 88000

gain_values = {
    GAIN_LOW: adafruit_tsl2591.GAIN_LOW,
    GAIN_MEDIUM: adafruit_tsl2591.GAIN_MED,
    GAIN_HIGH: adafruit_tsl2591.GAIN_HIGH,
    GAIN_MAX: adafruit_tsl2591.GAIN_MAX
}

integration_times = {
    IT_TIME_100MS: adafruit_tsl2591.INTEGRATIONTIME_100MS,
    IT_TIME_200MS: adafruit_tsl2591.INTEGRATIONTIME_200MS,
    IT_TIME_300MS: adafruit_tsl2591.INTEGRATIONTIME_300MS,
    IT_TIME_400MS: adafruit_tsl2591.INTEGRATIONTIME_400MS,
    IT_TIME_500MS: adafruit_tsl2591.INTEGRATIONTIME_500MS,
    IT_TIME_600MS: adafruit_tsl2591.INTEGRATIONTIME_600MS
}


class TSL2591(Sensor):
    """Concrete implementation of the TSL2591 sensor."""

    @property
    def name(self):
        return SENSOR_TSL2591

    def read(self):
        """Reads the full spectrum light in lux, and raw human visible light plus raw IR light."""
        logging.info('Reading TSL2591 sensor')
        sensor = adafruit_tsl2591.TSL2591(self._i2c)
        if tsl2591_gain in gain_values.keys():
            sensor.gain = gain_values[tsl2591_gain]
            logging.debug('set gain to %s', tsl2591_gain)
        else:
            logging.warning('Gain %s is not supported', tsl2591_gain)

        if tsl2591_integration_time in integration_times.keys():
            sensor.integration_time = integration_times[tsl2591_integration_time]
            logging.debug('set integration_time to %s', tsl2591_integration_time)
        else:
            logging.warning('Integration time %s is not supported', tsl2591_integration_time)

        # Read lux property of sensor. Can raise RuntimeError in very bright circumstances.
        try:
            lux = sensor.lux
        except RuntimeError:
            logging.warning("Sensor overflow detected while reading lux")
            lux = _MAX_LUX
        measurements = {
            'lux': lux,
            'visible-light': sensor.visible,
            'IR-light': sensor.infrared
        }
        logging.info('Measurement: %s', measurements)
        return measurements



#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.

"""
Utility to interact with the /stations endpoint
"""
import logging

import requests

from settings import timeout, verify, station_name, stations_endpoint, sensors


def get_sensors():
    """Retrieve the names of the sensors as comma separated string for the station"""
    logging.debug('Retrieving sensors for this station')
    try:
        uri = f'{stations_endpoint}/{station_name}'
        response = requests.get(uri, verify=verify, timeout=timeout)
        data = response.json()['sensors']
        logging.debug('Response body: %s', response.json())
        _sensors = ','.join(map(lambda sensor: sensor['name'], data))
    except Exception as exception:
        logging.exception('Failed to retrieve sensors for station from server', exc_info=exception)
        logging.warning('Falling back to locally configured sensors')
        _sensors = sensors

    logging.debug('Sensors: %s', _sensors)
    return _sensors


def get_sensor_settings(sensor_name):
    """Retrieve the settings of a single sensor for this station"""
    logging.debug('Retrieving sensor settings for %s', sensor_name)
    try:
        uri = f'{stations_endpoint}/{station_name}/sensors/{sensor_name}/settings'
        response = requests.get(uri, verify=verify, timeout=timeout)
        logging.debug('Response body: %s', response.json())
        return response.json()['settings']
    except Exception as exception:
        logging.exception('Failed to retrieve sensor settings', exc_info=exception)
        return {}



#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.
"""The abstract base class (ABC) for every sensor."""

import logging
from abc import ABCMeta, abstractmethod

from settings import disable_I2C
from station import get_sensor_settings

if not disable_I2C:
    import busio
    from board import SCL, SDA


class Sensor(metaclass=ABCMeta):
    """
    A base class for sensors connected to the station via I2C.
    """

    @property
    @abstractmethod
    def name(self):
        """The name of the sensor, as known by the server"""
        raise NotImplementedError

    def __init__(self):
        """Initialize the I2C bus to communicate with the sensors. When in simulation mode, I2C is not initialized."""
        if not disable_I2C:
            self._i2c = busio.I2C(SCL, SDA)
        self._settings = get_sensor_settings(self.name)
        logging.debug('Sensor settings: %s', self._settings)

    @abstractmethod
    def read(self):
        """
        Connect to the sensor over I2C and perform a measurement and return the value(s) that were measured.
        May return multiple values if the sensor measures multiple metrics.

        :rtype dict
        """
        raise NotImplementedError
